module tomo_mod

implicit none

contains

!!---------------------------------------------------------- 

subroutine FirstDer1d(souIn, souOut)
  !! take 1st derivative
  integer                       :: n
  real, dimension(:)            :: souIn, souOut
  real, dimension(size(souIn))  :: temp

  n = size(souIn)
  temp = souIn
  souOut = 0.

  souOut(2:n) = temp(2:n) - temp(1:n-1)

end subroutine

!!---------------------------------------------------------- 

subroutine SecDer1d(souIn, souOut)
  !! take 2nd derivative
  integer                       :: n
  real, dimension(:)            :: souIn, souOut
  real, dimension(size(souIn))  :: temp

  n = size(souIn)
  temp = souIn
  souOut = 0.

  souOut(2:n-1) =  temp(1:n-2) - 2*temp(2:n-1) + temp(3:n)

end subroutine

!!---------------------------------------------------------- 

subroutine AbcWt(vel0e, vwt, wt, dt, d1, d2, bc1, bc2, bc3, bc4, fat)
  !! compute absorbing boundary condition weights
  integer                       :: bc1, bc2, bc3, bc4, fat
  real                          :: dt, d1, d2
  real, dimension(:,:)          :: vel0e, vwt, wt
  integer                       :: iz, ix, n1e, n2e

  n1e = size(vel0e,1)
  n2e = size(vel0e,2)
  if(maxval(vel0e)-minval(vel0e)>2.0)vel0e = 0.2*vel0e+2.2

  vwt = 0.
  wt = 0.

  do ix=1+fat,n2e-fat
    do iz=1+fat,bc1+fat
      vwt(iz,ix) = vel0e(iz,ix)*dt/d1
      wt(iz,ix) = wt(iz,ix) + float(bc1+fat-iz+1)/float(bc1)*.15*vwt(iz,ix)
    enddo
  enddo

  do ix=1+fat,n2e-fat
    do iz=n1e-bc3-fat+1,n1e-fat
      vwt(iz,ix) = vel0e(iz,ix)*dt/d1
      wt(iz,ix) = wt(iz,ix) + float(iz-n1e+bc3+fat)/float(bc3)*.15*vwt(iz,ix)
    enddo
  enddo

  do ix=1+fat,bc2+fat
    do iz=1+fat,n1e-fat
      vwt(iz,ix) = vel0e(iz,ix)*dt/d2
      wt(iz,ix) = wt(iz,ix) + float(bc2+fat-ix+1)/float(bc2)*.15*vwt(iz,ix)
    enddo
  enddo

  do ix=n2e-bc4-fat+1,n2e-fat
    do iz=1+fat,n1e-fat
      vwt(iz,ix) = vel0e(iz,ix)*dt/d2
      wt(iz,ix) = wt(iz,ix) + float(ix-n2e+bc4+fat)/float(bc4)*.15*vwt(iz,ix)
    enddo
  enddo

  call extend(vwt, fat, fat, fat, fat)

end subroutine

!!---------------------------------------------------------- 

subroutine extend(model, crop1, crop2, crop3, crop4)
  integer                       :: crop1, crop2, crop3, crop4
  real, dimension(:,:)          :: model
  integer                       :: iz, ix, n1e, n2e

  n1e = size(model,1)
  n2e = size(model,2)

  do iz=1,crop1
    model(iz,:) = model(1+crop1,:)
  enddo

  do iz=n1e-crop3+1,n1e
    model(iz,:) = model(n1e-crop3,:)
  enddo

  do ix=1,crop2
    model(:,ix) = model(:,1+crop2)
  enddo

  do ix=n2e-crop4+1,n2e
    model(:,ix) = model(:,n2e-crop4)
  enddo

end subroutine

!!---------------------------------------------------------- 

subroutine stepper(r0, r1, r2, r3, step_size, stat)
  !! testing step sizes
  double precision              :: r0, r1, r2, r3, step_size
  integer                       :: stat

  if(isnan(r1) .or. r1 .eq. r1 + r1) then
    write(0,*) "something wrong with r1, will reset its value"
    r1 = r0 + abs(r0)
  endif
  if(isnan(r2) .or. r2 .eq. r2 + r2) then
    write(0,*) "something wrong with r2, will reset its value"
    r2 = r0 + abs(r0)
  endif
  if(isnan(r3) .or. r3 .eq. r3 + r3) then
    write(0,*) "something wrong with r3, will reset its value"
    r3 = r0 + abs(r0)
  endif

  if(r0 < r1 .and. r0 < r2 .and. r0 < r3) then
    write(0,*) "all residuals are larger than r0, will try dividing by 4"
    step_size = step_size * 0.25
    stat = 1
  elseif(r1 < r2 .and. r1 < r3) then
    write(0,*) "step scalar is = 0.5"
    step_size = step_size * 0.5
    write(0,*) "The best step size is = ", step_size
    stat = 0
  elseif(r2 < r1 .and. r2 < r3) then
    write(0,*) "step scalar is = 1.0"
    write(0,*) "The best step size is = ", step_size
    stat = 0
  elseif(r3 < r1 .and. r3 < r2) then
    write(0,*) "step scalar is = 1.5"
    step_size = step_size * 1.5
    write(0,*) "The best step size is = ", step_size
    stat = 0
  else
    write(0,*) "unknown step size problem, will try dividing by 4"
    step_size = step_size * 0.25
    stat = 1
  endif

end subroutine

!!---------------------------------------------------------- 

subroutine hwindow(refl, hwt, h_nx, h_ox, eps)
  !! testing step sizes
  integer                       :: h_nx, h_ox
  real                          :: eps
  real, dimension(:,:,:)        :: refl, hwt
  integer                       :: ih, k
  real                          :: e0, e1

  if(h_ox .ne. -(h_nx-1.)/2) write(0,*) "h-axis is not symmetric, will not work properly"

  e0 = sqrt(sum(refl*refl)-sum(refl(:,:,1-h_ox)*refl(:,:,1-h_ox)))

  do k=1,h_nx/2
    e1 = sqrt(sum(refl(:,:,1:k)*refl(:,:,1:k))+sum(refl(:,:,h_nx-k+1:h_nx)*refl(:,:,h_nx-k+1:h_nx)))
    if(e1>e0/100.) exit
  end do

  if(k>1) then
    write(0,*) "h-axis is reduces with k=", k
    hwt = 0.
    do ih=k,h_nx-k+1
      hwt(:,:,ih) = ih-1+h_ox
    enddo
    hwt = abs(hwt)
    if(maxval(hwt) > 0.) hwt = eps*hwt/maxval(hwt)
    do ih=1,k-1
      hwt(:,:,ih) = eps
      hwt(:,:,h_nx-ih+1) = eps
    enddo
  else
    write(0,*) "h-axis will not be reduced"
  endif

end subroutine


end module
